#! /bin/sh

init() {
    tmpd=`mktemp -d ${TMPDIR:-/tmp}/err-names-XXXXXX`
    test -d "${tmpd}" || {
        tmpd=${TMPDIR:-/tmp}/err-names-$$
        rm -rf ${tmpd}
        mkdir ${tmpd}
    }
    tmpf=${tmpd}/temp.c
    trap "rm -rf ${tmpd}" EXIT
    echo '#include <errno.h>' > ${tmpf}

    cc=`command -v gcc 2>/dev/null`
    test -x "$cc" || cc=`command -v cc`

    errno_list=`
      $cc -E ${tmpf} | \
        sed -n -e 's/\.h[^-_a-zA-Z0-9/].*/.h/' \
            -e 's@.*\(/usr/include.*/errno\)@\1@p' | \
        sort -u `

    grep_expr='^[ 	]*#[ 	]*define[ 	]+E[A-Z_0-9]+[ 	]*[0-9]'
    sed_cmd='s/^[ 	]*#[ 	]*define[ 	]*//'

    egrep -h "$grep_expr" $errno_list | \
        sed "$sed_cmd" | \
        awk '{ print $1 " " $2 }' | \
        sort -u -n -k2 > ${tmpd}/sorted.txt
}

compute_limits() {
    max_val=`sed -n '$s/^[A-Z_0-9]*[ 	]*//p' ${tmpd}/sorted.txt`
    max_width=0
    last_val=-1
    offset=0

    exec 6> ${tmpd}/err-offsets 7> ${tmpd}/string-frags
    {
        while read nm val
        do
            nmwidth=`echo $nm | wc -c`
            if test $last_val -ne $val
            then
                last_val=`expr $last_val + 1`
                while test $last_val -lt $val
                do
                    echo -1 >&6
                    last_val=`expr $last_val + 1`
                done
                printf '"%s\\0"\n' ${nm} >&7
                echo $offset >&6
                nm=`echo $nm | wc -c`
                offset=`expr $offset + ${nmwidth}`
            fi

            test $nmwidth -gt $max_width && max_width=$nmwidth
        done

        echo $offset > ${tmpd}/offset
        expr $max_width - 1 > ${tmpd}/max_width
    } < ${tmpd}/sorted.txt

    max_width=`cat ${tmpd}/max_width`
    offset=`cat ${tmpd}/offset`
    exec 6>&- 7>&-
}

gen_header() {
    nm=`command -v columns` 2>/dev/null
    test -x $nm || {
        columns() {
            if test X$3 = X-S,
            then
                sed -e 's/^/    /' -e 's/$/,/' -e '$s/,/ };/' \
                    ${tmpd}/err-offsets
            else
                sed -e 's/^/    /' -e '$s/$/;/' \
                    ${tmpd}/string-frags
            fi
            echo
        }
    }

    static_pfx='static '
    guard=ERRNO_LIMIT
    max_val=`expr $max_val + 1`

    cat > ${1} <<- _EOF_
	/* -*- buffer-read-only: t -*- vi: set ro:
	 *
	 * DO NOT EDIT THIS FILE
	 *
	 * This file has been derived by scanning system headers
	 * on `date`.
	 *
	`echo "${errno_list}" | sed 's/^/ * /'`
	 *
	`sed -n '/\* *Copyright/,/^ \*\// {
		  p
		  /^ \*\//q
		}' errno.c`

	/**
	 * @file ${1}
	 *
	 * @brief
	 * The names, values and counts of error numbers.
	 *
	 * This has been tailored to the `uname -s` version
	 * `uname -r` system for the `uname -m` platform.
	 */

	#ifndef ${guard}
	/**
	 * A count of offset entries for the error name offset table.
	 * In other words, an upper limit on the value of error numbers.
	 * Also used as a header multi-inclusion guard.
	 */
	#define ${guard} ${max_val}

	/**
	 * The names of the error numbers known on this system.
	 */
	${static_pfx}char const errno_names[$offset] =
	`columns -I4 --spread=1 --end=";" -i ${tmpd}/string-frags`

	/**
	 * table of error name offsets
	 */
	${static_pfx}int const errno_offsets[${max_val}] = {
	`columns -I4 --spread=1 -S, --end=" };" -i ${tmpd}/err-offsets`

	/**
	 * error number display format.  It has been adjusted so that all
	 * the descriptions line up vertically.
	 * The longest error name is ${max_width} characters.
	 */
	${static_pfx}char const show_fmt[] = "%5d (%-${max_width}s) == %s\n";

	#endif /* ${guard} */
	/* end of ${1} */
	_EOF_
}

gen_docs() {
    nm=`command -v doxygen`
    test -x "$nm" || exit 0
    test -d doxygen || mkdir doxygen
    doxygen -g - | \
      sed -e '/^#/d' -e '/^[ 	]*$/d' \
	-e '/^FULL_PATH_NAMES /s/=.*/= NO/' \
	-e '/^OPTIMIZE_OUTPUT_FOR_C /s/=.*/= YES/' \
	-e '/^EXTRACT_/s/=.*/= YES/' \
	-e '/^SOURCE_BROWSER /s/=.*/= YES/' \
	-e '/^PROJECT_NAME /s/=.*/= errno/' \
	-e '/^PROJECT_BRIEF /s/=.*/= show error number descriptions/' \
	-e '/^INPUT /s/=.*/= errno.c err-names.h/' > doxygen/Doxyfile

    nm=`doxygen doxygen/Doxyfile 2>&1` || {
        err=$?
        exec 1>&2
        echo "doxygen exited with $err while producing docs"
        echo "$nm"
        exit 1
    }

    for f in *
    do
        test -d "$f" || continue
        case "$f" in
        doxygen ) : ;;
        * ) test -d doxygen/$f && rm -rf doxygen/$f
            mv -f $f doxygen/. ;;
        esac
    done

    test -d ~/public_html || return 0
    test -d doxygen/html  || return 0
    test -d ~/public_html/errno && rm -rf ~/public_html/errno
    cp -frp doxygen/html ~/public_html/errno
}

init
compute_limits
gen_header err-names.h
gen_docs
exit 0
# end of mk-err-names.sh
